﻿-- AddOn nameSpace
local SoDEC = LibStub("AceAddon-3.0"):NewAddon("SoDEndlessChat", "AceConsole-3.0", "AceHook-3.0")

-- UpValue
local _G = _G
local tonumber = _G.tonumber
local tostring = _G.tostring
local string_find = _G.string.find
local string_match = _G.string.match
local string_gmatch = _G.string.gmatch
local string_utf8sub = _G.string.utf8sub
local string_sub = _G.string.sub
local string_gsub = _G.string.gsub
local string_utf8len = _G.string.utf8len
local string_len = _G.string.len
local string_format = _G.string.format
local table_insert = _G.table.insert
local table_remove = _G.table.remove
local wipe = _G.wipe

local GetTime = _G.GetTime

-- Constants
local BN_QUEUED = "[SoDEC] ".._G.BN_WHISPER..": |cff20ff20%d |r".._G.QUEUED_STATUS_QUEUED
local MAX_LETTERS = 254											-- Blizzard Maximum Letters

-- Frames
local Frame_OnUpdate = _G.CreateFrame("frame")					-- OnUpdate Frame

-- Tables
local OutputBNConvoTB = {}										-- BN Convo Whispers
local OutputBNWhisperTB = {}									-- BN Regular Whispers
local MsgLinksTB = {}											-- Temp For Links
local WordsTB = {}												-- Temp For Words
local MsgsTB = {}												-- Temp For Messages
local BNTimeSentTB = {}											-- BN Whisper Timer For Message Limit (Convos have no such limit)

-- Variables
local timerBNWhisper = 0										-- Total OnUpdate Timer For BNWhispers
local timerBNConvo = 0											-- Total OnUpdate Timer For BNConvos
local queueNotify = false										-- Queued BN Whispers Notify

--**********************************************************************************************************************************************************
-------------------------------------------------------------- Initialisation Functions
--**********************************************************************************************************************************************************

function SoDEC:OnInitialize()
end
------------------------------------------------------------------------------------------------------------------------------------------------------------

function SoDEC:OnEnable()
	self:SecureHook("ChatEdit_UpdateHeader", "ChatEdit_Update")
	self:RawHook("SendChatMessage", true)
	self:RawHook("BNSendWhisper", true)
	self:RawHook("BNSendConversationMessage", true)
end
------------------------------------------------------------------------------------------------------------------------------------------------------------

function SoDEC:OnDisable()
end

--**********************************************************************************************************************************************************
------------------------------------------------------------------- Tool Functions
--**********************************************************************************************************************************************************

local function LF_onUpdate(self, elapsed)
	if OutputBNWhisperTB[1] or OutputBNConvoTB[1] then
		-- BN Whispers
		if OutputBNWhisperTB[1] then
			timerBNWhisper = timerBNWhisper + elapsed
			
			-- Lets do this
			if timerBNWhisper >= .05 then
				timerBNWhisper = 0
				
				-- Throttle check for 10 message limit
				if not BNTimeSentTB[10] or BNTimeSentTB[1] < GetTime() - 10 then
					_G.BNSendWhisper(OutputBNWhisperTB[1].BNID, OutputBNWhisperTB[1].Message)
					table_remove(OutputBNWhisperTB,1)
				elseif not queueNotify then
					queueNotify = true
					print(string_format(BN_QUEUED, #OutputBNWhisperTB))
				end
			end
		end
		
		-- BN Conversations
		if OutputBNConvoTB[1] then
			timerBNConvo = timerBNConvo + elapsed
			
			-- Lets do this
			if timerBNConvo >= .05 then
				timerBNConvo = 0
				_G.BNSendConversationMessage(OutputBNConvoTB[1].BNID, OutputBNConvoTB[1].Message)
				table_remove(OutputBNConvoTB,1)
			end
		end
	else
		-- Cancel onUpdate
		Frame_OnUpdate:SetScript("OnUpdate", nil)
		queueNotify = false
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------

local function LF_SplitIntoMsgs(text)
	-- Parse out links
	text = string_gsub(text, "(|H[^|]+|h[^|]+|h)", function(theLink)
				table_insert(MsgLinksTB, theLink)
				local str = tostring(#MsgLinksTB)
				local length = string_len(theLink) - 4
				while string_len(str) < length do
					str = "0"..str
				end
				return "\001\002"..str.."\003\004"
	end)
	
	-- Breakdown words
	for v in string_gmatch(text, "[^ ]+") do
		if string_len(v) > MAX_LETTERS then
			local splitWord, i = "", 1
			while v and string_len(v) > 0 do
				splitWord = string_utf8sub(v, 1, MAX_LETTERS)
				local subtract = 2
				while string_len(splitWord) > MAX_LETTERS do
					splitWord = string_utf8sub(v, 1, MAX_LETTERS - subtract)
					subtract = subtract + 2
				end
				v = string_utf8sub(v, string_utf8len(splitWord) + 1, string_utf8len(v))
				table_insert(WordsTB, splitWord) 
				i = i + 1
				if i > 5 then break end
			end
		else 
			table_insert(WordsTB, v) 
		end
	end
	
	-- Recombine
	local temp = ""
	for i=1, #WordsTB + 1 do 
		if WordsTB[i] and (string_len(temp) + string_len(WordsTB[i])) < (MAX_LETTERS) then
			temp = ( temp ~= "" and temp.." "..WordsTB[i] ) or WordsTB[i]
		else
			temp = string_gsub(temp, "\001\002%d+\003\004", function(link)
						local index = tonumber(string_match(link, "(%d+)"))
						return MsgLinksTB[index] or link
			end)
			if temp ~= "" then
				table_insert(MsgsTB, temp)
			end
			temp = ( WordsTB[i] and WordsTB[i] ) or ""
		end
	end
	wipe(WordsTB)
	wipe(MsgLinksTB)
end
------------------------------------------------------------------------------------------------------------------------------------------------------------

local function LF_SplitNewLines(text)
	-- Find new-line count and change them so if something breaks in the future there is no continues loop
	local numFound
	text, numFound = string_gsub(text, "\\n", "\$n")
	
	-- Breakdown paragraphs and send them to be processed
	local position = 1
	for i=1, numFound do
		local start, finish = string_find(text, "\$n", position)
		LF_SplitIntoMsgs(string_sub(text, position, start - 1))
		position = finish + 1
	end
	LF_SplitIntoMsgs(string_utf8sub(text, position, string_utf8len(text)))
end

--**********************************************************************************************************************************************************
------------------------------------------------------------------- Call Functions
--**********************************************************************************************************************************************************

function SoDEC:ChatEdit_Update(self)
	local activeWindow = _G.ChatEdit_GetActiveWindow()
	if not activeWindow then return end
	
	local chattype = self:GetAttribute("chatType")
	if chattype ~= "YELL" then
		-- No Yell Spam Allowed! If you change this your a hooker :)
		self:SetMaxLetters(0)
		self:SetMaxBytes(0)
	else
		self:SetMaxLetters(MAX_LETTERS)
		self:SetMaxBytes(256)
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------

function SoDEC:SendChatMessage(...)
	local text, chatType, language, channel = ...
	
	-- Catch new-line, mark as new paragraph
	local found = string_find(text, "\\n")
	
	-- Check
	if found or string_len(text) > MAX_LETTERS then
		if found then
			LF_SplitNewLines(text)
		else
			LF_SplitIntoMsgs(text)
		end
		
		-- Send whisper to ChatThrottleLib
		for i=1, #MsgsTB do 
			_G.ChatThrottleLib:SendChatMessage("ALERT", "SoDEC", MsgsTB[i], chatType, language, channel)
		end
		wipe(MsgsTB)
	else
		-- Send whisper to blizzard
		SoDEC.hooks.SendChatMessage(...)
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------

function SoDEC:BNSendWhisper(...)
	local BNETID, text = ...
	
	-- Catch new-line, mark as new paragraph
	local found = string_find(text, "\\n")
	
	-- Check
	if found or string_len(text) > MAX_LETTERS then
		if found then
			LF_SplitNewLines(text)
		else
			LF_SplitIntoMsgs(text)
		end
		
		-- Set OnUpdate so the messages sent get spaced out (helps keep received in right order, due to blizzard side issues)
		timerBNWhisper = ( OutputBNWhisperTB[1] and timerBNWhisper ) or 1
		for i=1, #MsgsTB do
			OutputBNWhisperTB[1 + #OutputBNWhisperTB] = {BNID = BNETID, Message = MsgsTB[i]}
		end
		wipe(MsgsTB)
		queueNotify = false
		Frame_OnUpdate:SetScript("OnUpdate", LF_onUpdate)
	else
		-- Delete send times
		local timeNow = GetTime()
		if BNTimeSentTB[10] and BNTimeSentTB[1] < timeNow - 10 then
			local i = 1
			while i <= #BNTimeSentTB do
				if BNTimeSentTB[i] < timeNow - 10 then
					table_remove(BNTimeSentTB, i)
				else
					i = i + 1
				end
			end
		end
		
		-- Check if they are trying to whisper while the limit is hit
		if BNTimeSentTB[10] then 
			timerBNWhisper = ( OutputBNWhisperTB[1] and timerBNWhisper ) or 1
			OutputBNWhisperTB[1 + #OutputBNWhisperTB] = {BNID = BNETID, Message = text}
			queueNotify = false
			Frame_OnUpdate:SetScript("OnUpdate", LF_onUpdate)
		else		
			-- Send whisper to blizzard and set time of message
			BNTimeSentTB[1 + #BNTimeSentTB] = timeNow
			SoDEC.hooks.BNSendWhisper(...)
		end
	end
end
------------------------------------------------------------------------------------------------------------------------------------------------------------

function SoDEC:BNSendConversationMessage(...)
	local BNETID, text = ...
	
	-- Catch new-line, mark as new paragraph
	local found = string_find(text, "\\n")
	
	-- Check
	if found or string_len(text) > MAX_LETTERS then
		if found then
			LF_SplitNewLines(text)
		else
			LF_SplitIntoMsgs(text)
		end
		
		-- Set OnUpdate so the messages sent get spaced out (helps keep received in right order, due to blizzard side issues)
		timerBNConvo = ( OutputBNConvoTB[1] and timerBNConvo ) or 1
		for i=1, #MsgsTB do
			OutputBNConvoTB[1 + #OutputBNConvoTB] = {BNID = BNETID, Message = MsgsTB[i]}
		end
		wipe(MsgsTB)
		Frame_OnUpdate:SetScript("OnUpdate", LF_onUpdate)
	else
		-- Send whisper to blizzard
		SoDEC.hooks.BNSendConversationMessage(...)
	end
end

--**********************************************************************************************************************************************************
------------------------------------------------------------------- Slash Command
--**********************************************************************************************************************************************************

SLASH_SODENDLESSCHAT1 = "/cmds"
local function LF_handler(text, editbox)
	for v in string_gmatch(text, "/[^/]+") do
		editbox:SetText(v)
		ChatEdit_SendText(editbox)
	end
end
SlashCmdList["SODENDLESSCHAT"] = LF_handler